/*
 * USB Driver for ALi m5602 based webcams
 *
 * Copyright (C) 2008 Erik Andrén
 * Copyright (C) 2007 Ilyes Gouta. Based on the m5603x Linux Driver Project.
 * Copyright (C) 2005 m5603x Linux Driver Project <m5602@x3ng.com.br>
 *
 * Portions of code to USB interface and ALi driver software,
 * Copyright (c) 2006 Willem Duinker
 * v4l2 interface modeled after the V4L2 driver
 * for SN9C10x PC Camera Controllers
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, version 2.
 *
 */

#ifndef M5602_BRIDGE_H_
#define M5602_BRIDGE_H_

#include <linux/vmalloc.h>
#include <linux/version.h>
#include <linux/usb.h>
#include <linux/mm.h>

#include <media/v4l2-dev.h>
#include <media/v4l2-common.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 27)
#include <media/v4l2-ioctl.h>
#endif

/*****************************************************************************/

/* Debug parameters */
#define DBG_INIT 0x1
#define DBG_PROBE 0x2
#define DBG_V4L2 0x4
#define DBG_TRACE 0x8
#define DBG_DATA 0x10
#define DBG_V4L2_CID 0x20

#define PDEBUG(level, fmt, args...) \
	do { \
		if (debug & level)     \
			info("[%s:%d] " fmt, __func__, __LINE__ , \
			## args); \
	} while (0)

/*****************************************************************************/

#define M5602_XB_SENSOR_TYPE		0x00
#define M5602_XB_SENSOR_CTRL		0x01
#define M5602_XB_LINE_OF_FRAME_H	0x02
#define M5602_XB_LINE_OF_FRAME_L	0x03
#define M5602_XB_PIX_OF_LINE_H		0x04
#define M5602_XB_PIX_OF_LINE_L		0x05
#define M5602_XB_VSYNC_PARA		0x06
#define M5602_XB_HSYNC_PARA		0x07
#define M5602_XB_TEST_MODE_1		0x08
#define M5602_XB_TEST_MODE_2		0x09
#define M5602_XB_SIG_INI		0x0a
#define M5602_XB_DS_PARA		0x0e
#define M5602_XB_TRIG_PARA		0x0f
#define M5602_XB_CLK_PD			0x10
#define M5602_XB_MCU_CLK_CTRL		0x12
#define M5602_XB_MCU_CLK_DIV		0x13
#define M5602_XB_SEN_CLK_CTRL		0x14
#define M5602_XB_SEN_CLK_DIV		0x15
#define M5602_XB_AUD_CLK_CTRL		0x16
#define M5602_XB_AUD_CLK_DIV		0x17
#define M5602_XB_DEVCTR1		0x41
#define M5602_XB_EPSETR0		0x42
#define M5602_XB_EPAFCTR		0x47
#define M5602_XB_EPBFCTR		0x49
#define M5602_XB_EPEFCTR		0x4f
#define M5602_XB_TEST_REG		0x53
#define M5602_XB_ALT2SIZE		0x54
#define M5602_XB_ALT3SIZE		0x55
#define M5602_XB_OBSFRAME		0x56
#define M5602_XB_PWR_CTL		0x59
#define M5602_XB_ADC_CTRL		0x60
#define M5602_XB_ADC_DATA		0x61
#define M5602_XB_MISC_CTRL		0x62
#define M5602_XB_SNAPSHOT		0x63
#define M5602_XB_SCRATCH_1		0x64
#define M5602_XB_SCRATCH_2		0x65
#define M5602_XB_SCRATCH_3		0x66
#define M5602_XB_SCRATCH_4		0x67
#define M5602_XB_I2C_CTRL		0x68
#define M5602_XB_I2C_CLK_DIV		0x69
#define M5602_XB_I2C_DEV_ADDR		0x6a
#define M5602_XB_I2C_REG_ADDR		0x6b
#define M5602_XB_I2C_DATA		0x6c
#define M5602_XB_I2C_STATUS		0x6d
#define M5602_XB_GPIO_DAT_H		0x70
#define M5602_XB_GPIO_DAT_L		0x71
#define M5602_XB_GPIO_DIR_H		0x72
#define M5602_XB_GPIO_DIR_L		0x73
#define M5602_XB_GPIO_EN_H		0x74
#define M5602_XB_GPIO_EN_L		0x75
#define M5602_XB_GPIO_DAT		0x76
#define M5602_XB_GPIO_DIR		0x77
#define M5602_XB_MISC_CTL		0x70

#define I2C_BUSY 0x80

/*****************************************************************************/

/* Driver info */
#define DRIVER_AUTHOR "ALi m5602 Linux Driver Project"
#define DRIVER_DESC "ALi m5602 webcam driver"

#define M5602_MAX_FRAMES	32
#define M5602_URBS		2
#define M5602_ISOC_PACKETS	14

#define M5602_URB_TIMEOUT	msecs_to_jiffies(2 * M5602_ISOC_PACKETS)
#define M5602_URB_MSG_TIMEOUT   5000
#define M5602_FRAME_TIMEOUT	2

/*****************************************************************************/

static DECLARE_RWSEM(m5602_disconnect);

/* A skeleton used for sending messages to the m5602 bridge */
static const unsigned char bridge_urb_skeleton[] = {
	0x13, 0x00, 0x81, 0x00
};

/* A skeleton used for sending messages to the sensor */
static const unsigned char sensor_urb_skeleton[] = {
	0x23, M5602_XB_GPIO_EN_H, 0x81, 0x06,
	0x23, M5602_XB_MISC_CTRL, 0x81, 0x80,
	0x13, M5602_XB_I2C_DEV_ADDR, 0x81, 0x00,
	0x13, M5602_XB_I2C_REG_ADDR, 0x81, 0x00,
	0x13, M5602_XB_I2C_DATA, 0x81, 0x00,
	0x13, M5602_XB_I2C_CTRL, 0x81, 0x11
};

/* Enumerates all the frame states */
enum m5602_frame_state {
	FRAME_UNUSED,   /* Unused (no MCAPTURE) */
	FRAME_QUEUED,   /* Ready to start grabbing */
	FRAME_GRABBING, /* In the process of being grabbed into */
	FRAME_DONE,     /* Finished grabbing, but not been synced yet */
	FRAME_ERROR     /* Something bad happened while processing */
};

/* Enumerates all the different stream states */
enum m5602_stream_state {
	STREAM_OFF,
	STREAM_INTERRUPT,
	STREAM_ON
};

enum m5602_dev_state {
	DEV_INITIALIZED = 0x1,
	DEV_DISCONNECTED = 0x2,
	DEV_MISCONFIGURED = 0x4
};

/* Frame structure */
struct m5602_frame_t {
	struct list_head frame;

	unsigned char *rawdata;
	int framenum;
	struct v4l2_buffer buf;
	enum m5602_frame_state state;
	unsigned long vma_use_count;
};

struct m5602_camera {
	/* The name of the m5602 camera */
	char *name;

	/* A pointer to the currently connected sensor */
	struct m5602_sensor *sensor;

	struct device dev;

	/* the v4l video device for this device */
	struct video_device *vdev;

	/* the usb device for this camera */
	struct usb_device *udev;

	/* the interface for this usb device */
	struct usb_interface *interface;

	/* A buffer used to store usb messages */
	u8 *control_buffer;

	/* the size of the INT receive buffer */
	size_t int_in_size;

	/* the address of the INT in endpoint */
	__u8 int_in_endpointAddr;

	/* A pointer array of URBs */
	struct urb *urb[M5602_URBS];

	/* the buffer to receive ISOC data */
	unsigned char *isoc_in_buffer[M5602_URBS];

	/* the size of the ISOC receive buffer */
	size_t isoc_in_size;

	/* the address of the ISOC in endpoint */
	__u8 isoc_in_endpointAddr;

	/* Pointer to the current frame and an array of frames */
	struct m5602_frame_t *frame_current, frame[M5602_MAX_FRAMES];

	/* The current frame's id, used to detect frame boundaries */
	u8 frame_id;

	/* The current frame count and the number of allocated frame buffers */
	u32 frame_count, nbuffers;

	/* Two linked lists, one for free frames,
	one for finished frames and one for frames awaiting post processing */
	struct list_head inqueue, outqueue;

	enum v4l2_memory io;
	enum m5602_stream_state stream;
	enum m5602_dev_state state;

	/* The number of current users */
	u8 users;

	/* Various mutexes, used throughout the driver */
	struct mutex dev_mutex, fileop_mutex;

	/* A lock to protect the various linked lists */
	spinlock_t queue_lock;

	struct kref kref;

	/* Various waitqueues used throughout the driver */
	wait_queue_head_t open, wait_frame, wait_stream;
};

int m5602_read_bridge(
	struct m5602_camera *cam, u8 address, u8 *i2c_data);

int m5602_write_bridge(
	struct m5602_camera *cam, u8 address, u8 i2c_data);

#endif
