/*
 * Driver for the po1030 sensor
 *
 * Copyright (c) 2008 Erik Andrén
 * Copyright (c) 2007 Ilyes Gouta. Based on the m5603x Linux Driver Project.
 * Copyright (c) 2005 m5603x Linux Driver Project <m5602@x3ng.com.br>
 *
 * Portions of code to USB interface and ALi driver software,
 * Copyright (c) 2006 Willem Duinker
 * v4l2 interface modeled after the V4L2 driver
 * for SN9C10x PC Camera Controllers
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, version 2.
 *
 */

#include "m5602_po1030.h"

int po1030_probe(struct m5602_camera *cam)
{
	u8 prod_id = 0, ver_id = 0, i;

	if (force_sensor) {
		if (force_sensor == PO1030_SENSOR) {
			info("Forcing a %s sensor", po1030.name);
			return 0;
		}
		/* If we want to force another sensor, don't try to probe this
		 * one */
		return -ENODEV;
	}

	info("Probing for a po1030 sensor");

	/* Run the pre-init to actually probe the unit */
	for (i = 0; i < ARRAY_SIZE(preinit_po1030); i++) {
		u8 data = preinit_po1030[i][2];
		if (preinit_po1030[i][0] == SENSOR)
			po1030_write_sensor(cam,
				preinit_po1030[i][1], &data, 1);
		else
			m5602_write_bridge(cam, preinit_po1030[i][1], data);
	}

	if (po1030_read_sensor(cam, 0x3, &prod_id, 1))
		return -ENODEV;

	if (po1030_read_sensor(cam, 0x4, &ver_id, 1))
		return -ENODEV;

	if ((prod_id == 0x02) && (ver_id == 0xef)) {
		info("Detected a po1030 sensor");
		return 0;
	}

	return -ENODEV;
}

int po1030_read_sensor(struct m5602_camera *cam, const u8 address,
				u8 *i2c_data, const u8 len)
{
	int err, i;

	do {
		err = m5602_read_bridge(cam, M5602_XB_I2C_STATUS, i2c_data);
	} while ((*i2c_data & I2C_BUSY) && !err);

	m5602_write_bridge(cam, M5602_XB_I2C_DEV_ADDR,
			   cam->sensor->i2c_slave_id);
	m5602_write_bridge(cam, M5602_XB_I2C_REG_ADDR, address);
	m5602_write_bridge(cam, M5602_XB_I2C_CTRL, 0x10 + len);
	m5602_write_bridge(cam, M5602_XB_I2C_CTRL, 0x08);

	for (i = 0; i < len; i++) {
		err = m5602_read_bridge(cam, M5602_XB_I2C_DATA, &(i2c_data[i]));

		PDEBUG(DBG_TRACE, "Reading sensor register "
				"0x%x containing 0x%x ", address, *i2c_data);
	}
	return (err < 0) ? err : 0;
}

int po1030_write_sensor(struct m5602_camera *cam, const u8 address,
				u8 *i2c_data, const u8 len)
{
	int err, i;
	u8 *p;

	/* The po1030 only supports one byte writes */
	if (len > 1 || !len)
		return -EINVAL;

	memcpy(cam->control_buffer, sensor_urb_skeleton,
				sizeof(sensor_urb_skeleton));

	cam->control_buffer[11] = cam->sensor->i2c_slave_id;
	cam->control_buffer[15] = address;

	p = cam->control_buffer + 16;

	/* Copy a four byte write sequence for each byte to be written to */
	for (i = 0; i < len; i++) {
		memcpy(p, sensor_urb_skeleton + 16, 4);
		p[3] = i2c_data[i];
		p += 4;
		PDEBUG(DBG_TRACE, "Writing sensor register 0x%x with 0x%x",
		       address, i2c_data[i]);
	}

	/* Copy the footer */
	memcpy(p, sensor_urb_skeleton + 20, 4);

	/* Set the total length */
	p[3] = 0x10 + len;

	err = usb_control_msg(cam->udev, usb_sndctrlpipe(cam->udev, 0),
			      0x04, 0x40, 0x19,
			      0x0000, cam->control_buffer,
			      20 + len * 4, M5602_URB_MSG_TIMEOUT);

	return (err < 0) ? err : 0;
}

int po1030_init(struct m5602_camera *cam)
{
	int i, err = 0;

	/* Init the sensor */
	for (i = 0; i < ARRAY_SIZE(init_po1030); i++) {
		u8 data[2] = {0x00, 0x00};

		switch (init_po1030[i][0]) {
		case BRIDGE:
			err = m5602_write_bridge(cam,
				init_po1030[i][1],
				init_po1030[i][2]);
			break;

		case SENSOR:
			data[0] = init_po1030[i][2];
			err = po1030_write_sensor(cam,
				init_po1030[i][1], data, 1);
			break;

		case SENSOR_LONG:
			data[0] = init_po1030[i][2];
			data[1] = init_po1030[i][3];
			err = po1030_write_sensor(cam,
				init_po1030[i][1], data, 2);
			break;
		default:
			info("Invalid stream command, exiting init");
			return -EINVAL;
		}
	}

	if (dump_sensor)
		po1030_dump_registers(cam);

	return (err < 0) ? err : 0;
}

int po1030_power_down(struct m5602_camera *cam)
{
	return 0;
}

int po1030_get_ctrl(struct m5602_camera *cam,
				 struct v4l2_control *ctrl)
{
	u8 i2c_data;
	int err = 0;

	switch (ctrl->id) {
	case V4L2_CID_GAIN:
		err = po1030_read_sensor(cam, PO1030_REG_GLOBALGAIN,
					       &i2c_data, 1);
		ctrl->value = i2c_data;
		PDEBUG(DBG_V4L2_CID, "Read global gain %d", ctrl->value);
		break;

	case V4L2_CID_EXPOSURE:
		err = po1030_read_sensor(cam, PO1030_REG_INTEGLINES_H,
					       &i2c_data, 1);
		ctrl->value = (i2c_data << 8);

		err = po1030_read_sensor(cam, PO1030_REG_INTEGLINES_M,
					       &i2c_data, 1);
		ctrl->value |= i2c_data;

		PDEBUG(DBG_V4L2_CID, "Exposure read as %d", ctrl->value);
		break;

	case V4L2_CID_RED_BALANCE:
		err = po1030_read_sensor(cam, PO1030_REG_RED_GAIN,
					 &i2c_data, 1);
		ctrl->value = i2c_data;
		PDEBUG(DBG_V4L2_CID, "Read red gain %d", ctrl->value);
		break;

	case V4L2_CID_BLUE_BALANCE:
		err = po1030_read_sensor(cam, PO1030_REG_BLUE_GAIN,
				&i2c_data, 1);
		ctrl->value = i2c_data;
		PDEBUG(DBG_V4L2_CID, "Read blue gain %d", ctrl->value);
		break;

	default:
		PDEBUG(DBG_V4L2, "Unknown V4L2 CID, 0x%x", ctrl->value);
		return -EINVAL;
	}

	return 0;
}

int po1030_set_ctrl(struct m5602_camera *cam,
			   const struct v4l2_control *ctrl)
{
	u8 reg_data;
	int err = 0;

	switch (ctrl->id) {
	case V4L2_CID_GAIN:
		reg_data = ctrl->value & 0xff;
		PDEBUG(DBG_V4L2, "Set global gain to %d", reg_data);
		err = po1030_write_sensor(cam, PO1030_REG_GLOBALGAIN,
						&reg_data, 1);
		break;

	case V4L2_CID_EXPOSURE:
		PDEBUG(DBG_V4L2, "Set exposure to %d", ctrl->value & 0xffff);

		reg_data = ((ctrl->value & 0xff00) >> 8);
		PDEBUG(DBG_V4L2, "Set exposure to high byte to 0x%x",
		       reg_data);

		err = po1030_write_sensor(cam, PO1030_REG_INTEGLINES_H,
			&reg_data, 1);

		reg_data = (ctrl->value & 0xff);
		PDEBUG(DBG_V4L2, "Set exposure to low byte to 0x%x",
		       reg_data);
		err = po1030_write_sensor(cam, PO1030_REG_INTEGLINES_M,
								&reg_data, 1);
		break;

	case V4L2_CID_RED_BALANCE:
		reg_data = ctrl->value & 0xff;
		PDEBUG(DBG_V4L2, "Set red gain to %d", reg_data);
		err = po1030_write_sensor(cam, PO1030_REG_RED_GAIN,
					  &reg_data, 1);
		break;

	case V4L2_CID_BLUE_BALANCE:
		reg_data = ctrl->value & 0xff;
		PDEBUG(DBG_V4L2, "Set blue gain to %d", reg_data);
		err = po1030_write_sensor(cam, PO1030_REG_BLUE_GAIN,
					  &reg_data, 1);
		break;


	default:
		PDEBUG(DBG_V4L2, "Illegal V4L id %d, value %d",
		       ctrl->id, ctrl->value);
		return -EINVAL;
	}

	return (err < 0) ? err : 0;;
}

void po1030_dump_registers(struct m5602_camera *cam)
{
	int address;
	u8 value = 0;

	info("Dumping the po1030 sensor core registers");
	for (address = 0; address < 0x7f; address++) {
		po1030_read_sensor(cam, address, &value, 1);
		info("register 0x%x contains 0x%x",
		     address, value);
	}

	info("po1030 register state dump complete");

	info("Probing for which registers that are read/write");
	for (address = 0; address < 0xff; address++) {
		u8 old_value, ctrl_value;
		u8 test_value[2] = {0xff, 0xff};

		po1030_read_sensor(cam, address, &old_value, 1);
		po1030_write_sensor(cam, address, test_value, 1);
		po1030_read_sensor(cam, address, &ctrl_value, 1);

		if (ctrl_value == test_value[0])
			info("register 0x%x is writeable", address);
		else
			info("register 0x%x is read only", address);

		/* Restore original value */
		po1030_write_sensor(cam, address, &old_value, 1);
	}
}
