/*
 * Driver for the s5k83a sensor
 *
 * Copyright (C) 2008 Erik Andrén
 * Copyright (C) 2007 Ilyes Gouta. Based on the m5603x Linux Driver Project.
 * Copyright (C) 2005 m5603x Linux Driver Project <m5602@x3ng.com.br>
 *
 * Portions of code to USB interface and ALi driver software,
 * Copyright (c) 2006 Willem Duinker
 * v4l2 interface modeled after the V4L2 driver
 * for SN9C10x PC Camera Controllers
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, version 2.
 *
 */

#include <linux/kthread.h>
#include "m5602_s5k83a.h"

int s5k83a_probe(struct m5602_camera *cam)
{
	u8 prod_id = 0, ver_id = 0;
	int i, err = 0;

	if (force_sensor) {
		if (force_sensor == S5K83A_SENSOR) {
			info("Forcing a %s sensor", s5k83a.name);
			return 0;
		}
		/* If we want to force another sensor, don't try to probe this
		 * one */
		return -ENODEV;
	}

	info("Probing for a s5k83a sensor");

	/* Preinit the sensor */
	for (i = 0; i < ARRAY_SIZE(preinit_s5k83a) && !err; i++) {
		u8 data[2] = {preinit_s5k83a[i][2], preinit_s5k83a[i][3]};
		if (preinit_s5k83a[i][0] == SENSOR)
			err = s5k83a_write_sensor(cam, preinit_s5k83a[i][1],
				data, 2);
		else
			err = m5602_write_bridge(cam, preinit_s5k83a[i][1],
				data[0]);
	}

	/* We don't know what register (if any) that contain the product id
	 * Just pick the first addresses that seem to produce the same results
	 * on multiple machines */
	if (s5k83a_read_sensor(cam, 0x00, &prod_id, 1))
		return -ENODEV;

	if (s5k83a_read_sensor(cam, 0x01, &ver_id, 1))
		return -ENODEV;

	if ((prod_id == 0xff) || (ver_id == 0xff))
		return -ENODEV;
	else
		info("Detected a s5k83a sensor");

	return 0;
}

int s5k83a_read_sensor(struct m5602_camera *cam, const u8 address,
			      u8 *i2c_data, const u8 len)
{
	int err, i;

	do {
		err = m5602_read_bridge(cam, M5602_XB_I2C_STATUS, i2c_data);
	} while ((*i2c_data & I2C_BUSY) && !err);

	m5602_write_bridge(cam, M5602_XB_I2C_DEV_ADDR,
			   cam->sensor->i2c_slave_id);
	m5602_write_bridge(cam, M5602_XB_I2C_REG_ADDR, address);
	m5602_write_bridge(cam, M5602_XB_I2C_CTRL, 0x18 + len);

	do {
		err = m5602_read_bridge(cam, M5602_XB_I2C_STATUS, i2c_data);
	} while ((*i2c_data & I2C_BUSY) && !err);

	for (i = 0; i < len; i++) {
		err = m5602_read_bridge(cam, M5602_XB_I2C_DATA, &(i2c_data[i]));

		PDEBUG(DBG_TRACE, "Reading sensor register "
				  "0x%x containing 0x%x ", address, *i2c_data);
	}
	return (err < 0) ? err : 0;
}

int s5k83a_write_sensor(struct m5602_camera *cam, const u8 address,
			       u8 *i2c_data, const u8 len)
{
	int err, i;
	u8 *p;

	/* No sensor with a data width larger than 16 bits has yet been seen */
	if (len > 2 || !len)
		return -EINVAL;

	memcpy(cam->control_buffer, sensor_urb_skeleton,
	       sizeof(sensor_urb_skeleton));

	cam->control_buffer[11] = cam->sensor->i2c_slave_id;
	cam->control_buffer[15] = address;

	/* Special case larger sensor writes */
	p = cam->control_buffer + 16;

	/* Copy a four byte write sequence for each byte to be written to */
	for (i = 0; i < len; i++) {
		memcpy(p, sensor_urb_skeleton + 16, 4);
		p[3] = i2c_data[i];
		p += 4;
		PDEBUG(DBG_TRACE, "Writing sensor register 0x%x with 0x%x",
		       address, i2c_data[i]);
	}

	/* Copy the tailer */
	memcpy(p, sensor_urb_skeleton + 20, 4);

	/* Set the total length */
	p[3] = 0x10 + len;

	err = usb_control_msg(cam->udev, usb_sndctrlpipe(cam->udev, 0),
			      0x04, 0x40, 0x19,
			      0x0000, cam->control_buffer,
			      20 + len * 4, M5602_URB_MSG_TIMEOUT);

	return (err < 0) ? err : 0;
}

int s5k83a_init(struct m5602_camera *cam)
{
	int i, err = 0;
	struct task_struct *rotation_thread;

	for (i = 0; i < ARRAY_SIZE(init_s5k83a) && !err; i++) {
		u8 data[2] = {0x00, 0x00};

		switch (init_s5k83a[i][0]) {
		case BRIDGE:
			err = m5602_write_bridge(cam,
					init_s5k83a[i][1],
					init_s5k83a[i][2]);
			break;

		case SENSOR:
			data[0] = init_s5k83a[i][2];
			err = s5k83a_write_sensor(cam,
				init_s5k83a[i][1], data, 1);
			break;

		case SENSOR_LONG:
			data[0] = init_s5k83a[i][2];
			data[1] = init_s5k83a[i][3];
			err = s5k83a_write_sensor(cam,
				init_s5k83a[i][1], data, 2);
			break;
		default:
			info("Invalid stream command, exiting init");
			return -EINVAL;
		}
	}

	if (dump_sensor)
		s5k83a_dump_registers(cam);
	
	rotation_thread = kthread_create(rotation_thread_function, cam, "rotation thread");
	cam->sensor->private_data = (void *)rotation_thread;
	wake_up_process(rotation_thread);
	
	return (err < 0) ? err : 0;
}

int rotation_thread_function(void *data){
	struct m5602_camera *cam = (struct m5602_camera *)data;
	u8 reg, previous_rotation = 0;
		
	set_current_state(TASK_INTERRUPTIBLE);
	while(!schedule_timeout(500)){
		
		if (mutex_lock_interruptible(&cam->fileop_mutex))
			continue;
		
		if (cam->state & DEV_DISCONNECTED) {
			mutex_unlock(&cam->fileop_mutex);
			continue;
		}

		if (cam->state & DEV_MISCONFIGURED) {
			mutex_unlock(&cam->fileop_mutex);
			continue;
		}
		
		s5k83a_get_rotation(cam, &reg);
		if (previous_rotation != reg){
			previous_rotation = reg;
			info("Camera was flipped!");
			
			s5k83a_get_vflip(cam, &reg);
			reg = reg ? 0: 1;
			s5k83a_set_vflip(cam, reg);
			
			s5k83a_get_hflip(cam, &reg);
			reg = reg ? 0: 1;
			s5k83a_set_hflip(cam, reg);
		}
		
		mutex_unlock(&cam->fileop_mutex);
		set_current_state(TASK_INTERRUPTIBLE);
	}
	info("End of rotation thread!");
	return 0;
}

int s5k83a_power_down(struct m5602_camera *cam)
{
	if (cam->sensor->private_data){
		kthread_stop((struct task_struct *)cam->sensor->private_data);
	}
	return 0;
}

int s5k83a_get_ctrl(struct m5602_camera *cam, struct v4l2_control *ctrl)
{
	int err = 0;
	u8 reg_data;

	switch (ctrl->id) {
	case V4L2_CID_BRIGHTNESS:
		err = s5k83a_get_brightness(cam, &reg_data);
		ctrl->value = reg_data;
		PDEBUG(DBG_V4L2_CID, "Get brightness 0x%x", ctrl->value);
		break;

	case V4L2_CID_WHITENESS:
		err = s5k83a_get_whiteness(cam, &reg_data);
		ctrl->value = reg_data;
		PDEBUG(DBG_V4L2_CID, "Get whiteness 0x%x", ctrl->value);
		break;

	case V4L2_CID_GAIN:
		err = s5k83a_get_gain(cam, &reg_data);
		ctrl->value = reg_data;
		PDEBUG(DBG_V4L2_CID, "Get gain 0x%x", ctrl->value);
		break;
		
	case V4L2_CID_HFLIP:
		err = s5k83a_get_hflip(cam, &reg_data);
		ctrl->value = reg_data;
		PDEBUG(DBG_V4L2_CID, "Get hflip 0x%x", ctrl->value);
		break;		
	
	case V4L2_CID_VFLIP:
		err = s5k83a_get_vflip(cam, &reg_data);
		ctrl->value = reg_data;
		PDEBUG(DBG_V4L2_CID, "Get vflip 0x%x", ctrl->value);
		break;		
		
	case S5K83A_PRIVATE_CID_ROTATION:
		err = s5k83a_get_rotation(cam, &reg_data);
		ctrl->value = reg_data;
		PDEBUG(DBG_V4L2_CID, "Get rotation 0x%x", ctrl->value);
		break;
		
	default:
		PDEBUG(DBG_V4L2_CID, "Unknown V4L2 CID 0x%x", ctrl->id);
		err = -EINVAL;
	}
	return (err < 0) ? err : 0;
}

int s5k83a_set_ctrl(struct m5602_camera *cam,
			   const struct v4l2_control *ctrl)
{
	int err = 0;

	switch (ctrl->id) {
	case V4L2_CID_BRIGHTNESS:
		err = s5k83a_set_brightness(cam, ctrl->value);
		PDEBUG(DBG_V4L2_CID, "Set brightness to 0x%x", ctrl->value);
		break;

	case V4L2_CID_WHITENESS:
		err = s5k83a_set_whiteness(cam, ctrl->value);
		PDEBUG(DBG_V4L2_CID, "Set whiteness to 0x%x", ctrl->value);
		break;

	case V4L2_CID_GAIN:
		err = s5k83a_set_gain(cam, ctrl->value);
		PDEBUG(DBG_V4L2_CID, "Set gain to 0x%x", ctrl->value);
		break;

	case V4L2_CID_HFLIP:
		err = s5k83a_set_hflip(cam, ctrl->value);
		PDEBUG(DBG_V4L2_CID, "Set hflip 0x%x", ctrl->value);
		break;		

	case V4L2_CID_VFLIP:
		err = s5k83a_set_vflip(cam, ctrl->value);
		PDEBUG(DBG_V4L2_CID, "Set vflip 0x%x", ctrl->value);
		break;		
	
	case S5K83A_PRIVATE_CID_ROTATION:
		err = -EINVAL;
		PDEBUG(DBG_V4L2_CID, "Get rotation is only for read!");
		break;
		
	default:
		PDEBUG(DBG_V4L2_CID,
		       "Illegal V4L2 CID %d, value %d", ctrl->id, ctrl->value);
		err = -EINVAL;
	}
	return (err < 0) ? err : 0;
}

void s5k83a_dump_registers(struct m5602_camera *cam)
{
	int address;
	u8 page, old_page;
	s5k83a_read_sensor(cam, S5K83A_PAGE_MAP, &old_page, 1);
	for (page = 0; page < 16; page++) {
		s5k83a_write_sensor(cam, S5K83A_PAGE_MAP, &page, 1);
		info("Dumping the s5k83a register state for page 0x%x", page);
		for (address = 0; address <= 0xff; address++) {
			u8 value = 0;
			s5k83a_read_sensor(cam, address, &value, 1);
			info("register 0x%x contains 0x%x",
			     address, value);
		}
	}
	info("s5k83a register state dump complete");

	for (page = 0; page < 16; page++) {
		s5k83a_write_sensor(cam, S5K83A_PAGE_MAP, &page, 1);
		info("Probing for which registers that are read/write "
		      "for page 0x%x", page);
		for (address = 0; address <= 0xff; address++) {
			u8 old_value, ctrl_value, test_value = 0xff;

			s5k83a_read_sensor(cam, address, &old_value, 1);
			s5k83a_write_sensor(cam, address, &test_value, 1);
			s5k83a_read_sensor(cam, address, &ctrl_value, 1);

			if (ctrl_value == test_value)
				info("register 0x%x is writeable", address);
			else
				info("register 0x%x is read only", address);

			/* Restore original value */
			s5k83a_write_sensor(cam, address, &old_value, 1);
		}
	}
	info("Read/write register probing complete");
	s5k83a_write_sensor(cam, S5K83A_PAGE_MAP, &old_page, 1);
}

int s5k83a_get_brightness(struct m5602_camera *cam, u8 *reg_data)
{
	int err = 0;
	u8 data[2];

	err = s5k83a_read_sensor(cam, S5K83A_BRIGHTNESS, data, 2);
	data[1] = data[1] << 1;
	*reg_data = data[1];

	return (err < 0) ? err : 0;
}

int s5k83a_set_brightness(struct m5602_camera *cam, u8 value)
{
	int err = 0;
	u8 data[2];

	data[0] = 0x00;
	data[1] = 0x20;
	err = s5k83a_write_sensor(cam, 0x14, data, 2);
	if (err < 0)
		return err;

	data[0] = 0x01;
	data[1] = 0x00;
	err = s5k83a_write_sensor(cam, 0x0d, data, 2);
	if (err < 0)
		return err;

	/* FIXME: This is not sane, we need to figure out the composition
		  of these registers */
	data[0] = value >> 3; /* brightness, high 5 bits */
	data[1] = value >> 1; /* brightness, high 7 bits */
	err = s5k83a_write_sensor(cam, S5K83A_BRIGHTNESS, data, 2);

	return (err < 0) ? err : 0;
}

int s5k83a_get_whiteness(struct m5602_camera *cam, u8 *reg_data)
{
	int err = 0;
	u8 data;

	err = s5k83a_read_sensor(cam, S5K83A_WHITENESS, &data, 1);

	*reg_data = data;
	return (err < 0) ? err : 0;
}

int s5k83a_set_whiteness(struct m5602_camera *cam, u8 value)
{
	int err = 0;
	u8 data[1];

	data[0] = value;
	err = s5k83a_write_sensor(cam, S5K83A_WHITENESS, data, 1);

	return (err < 0) ? err : 0;
}

int s5k83a_get_gain(struct m5602_camera *cam, u8 *reg_data)
{
	int err = 0;
	u8 data[2];

	err = s5k83a_read_sensor(cam, S5K83A_GAIN, data, 2);

	data[1] = data[1] & 0x3f;
	if (data[1] > S5K83A_MAXIMUM_GAIN)
		data[1] = S5K83A_MAXIMUM_GAIN;

	*reg_data = data[1];

	return (err < 0) ? err : 0;
}

int s5k83a_set_gain(struct m5602_camera *cam, u8 value)
{
	int err = 0;
	u8 data[2];

	data[0] = 0;
	data[1] = value;
	err = s5k83a_write_sensor(cam, S5K83A_GAIN, data, 2);

	return (err < 0) ? err : 0;
}

int s5k83a_set_hflip(struct m5602_camera *cam, u8 value){
	int err = 0;
	u8 data[1];

	data[0] = 0x05;
	err = s5k83a_write_sensor(cam, S5K83A_PAGE_MAP, data, 1);
	if (err < 0)
		return err;

	err = s5k83a_get_rotation(cam, &data[0]);
	if (data[0])
		value = value ? 0: 1;
	
	err = s5k83a_read_sensor(cam, S5K83A_FLIP, data, 1);
	if (err < 0)
		return err;
	
	/* set or zero seven bit, six bit is vflip */
	data[0] = value ? (data[0] & 0x40) | 0x80 | S5K83A_FLIP_MASK: (data[0] & 0x40) | S5K83A_FLIP_MASK;
	err = s5k83a_write_sensor(cam, S5K83A_FLIP, data, 1);
	if (err < 0)
		return err;
	
	data[0] = value ? 0x0a : 0x0b;
	err = s5k83a_write_sensor(cam, S5K83A_HFLIP_TUNE, data, 1);
	
	return (err < 0) ? err : 0;	
}

int s5k83a_get_hflip(struct m5602_camera *cam, u8 *reg_data){
	int err = 0;
	u8 data[1];

	data[0] = 0x05;
	err = s5k83a_write_sensor(cam, S5K83A_PAGE_MAP, data, 1);
	if (err < 0)
		return err;

	err = s5k83a_read_sensor(cam, S5K83A_FLIP, data, 1);
	*reg_data = (data[0] & 0x80)? 1 : 0;
	
	err = s5k83a_get_rotation(cam, &data[0]);
	if (data[0])
		*reg_data = *reg_data ? 0 : 1;
	
	return (err < 0) ? err : 0;	
}

int s5k83a_set_vflip(struct m5602_camera *cam, u8 value){
	int err = 0;
	u8 data[1];

	data[0] = 0x05;
	err = s5k83a_write_sensor(cam, S5K83A_PAGE_MAP, data, 1);
	if (err < 0)
		return err;
	
	err = s5k83a_get_rotation(cam, &data[0]);
	if (data[0])
		value = value ? 0: 1;	

	err = s5k83a_read_sensor(cam, S5K83A_FLIP, data, 1);
	if (err < 0)
		return err;
	
	/* set or zero six bit, seven is hflip */
	data[0] = value ? (data[0] & 0x80) | 0x40 | S5K83A_FLIP_MASK: (data[0] & 0x80) | S5K83A_FLIP_MASK;
	err = s5k83a_write_sensor(cam, S5K83A_FLIP, data, 1);
	if (err < 0)
		return err;
	
	data[0] = value? 0x0b : 0x0a;
	err = s5k83a_write_sensor(cam, S5K83A_VFLIP_TUNE, data, 1);
	
	return (err < 0) ? err : 0;	
}

int s5k83a_get_vflip(struct m5602_camera *cam, u8 *reg_data){
	int err = 0;
	u8 data[1];

	data[0] = 0x05;
	err = s5k83a_write_sensor(cam, S5K83A_PAGE_MAP, data, 1);
	if (err < 0)
		return err;

	err = s5k83a_read_sensor(cam, S5K83A_FLIP, data, 1);
	*reg_data = (data[0] & 0x40)? 1 : 0;
	
	err = s5k83a_get_rotation(cam, &data[0]);
	if (data[0])
		*reg_data = *reg_data ? 0 : 1;
	
	return (err < 0) ? err : 0;	
}

int s5k83a_get_rotation(struct m5602_camera *cam, u8 *reg_data){
	/* Get camera rotation on Acer notebooks */
	int err = 0;	
	err = m5602_read_bridge(cam, M5602_XB_GPIO_DAT, reg_data);
	*reg_data = (*reg_data & 0x40)? 0 : 1;
	return err;
}


